;compiler: avrasm
	.include "m32def.inc"
	.include "macros.asm"
	
	.equ	XTAL=16000000	;clock freq in Hz
	
	
	.cseg
	.org	0x0000
	rjmp	Init	;jump to demo code
	.org	0x0008
	rjmp	INTLCDDMArefresh	;LCD DMA refresh INT (timer 2 CTC)


;========================================================================================================================
;========================================================================================================================
;	init, demo code
;========================================================================================================================
;========================================================================================================================
Init:	outi	ddrc,0b00001111	;LCD data pins
	outi	ddrd,(1<<LCK)|(1<<LRS)	;LCD ctrl
	
	cbi	portc,LCK	;LCD E low
	
	outiw	sp,ramend	;stack to RAM end
	
	
	;USART init as 19200bd RX,TX
	outi	ucsra,(1<<u2x)
	outi	ucsrb,(1<<txen)|(1<<rxen)
	outiw	ubrr,XTAL/8/19200
	
	
	ldi	R16,2
	rcall	DelayVar	;wait 20ms
	
	sei		;enable interruption
	
	ldi	R16,2	;2 lines
	rcall	Dinit	;init DMA LCD refresh
	
	
	;some text	
	ldiw	R30,Text1T*2
	rcall	DwrStrFLASH
	rcall	Drefresh	;manual start LCD refresh
	
	ldi	R16,200
	rcall	DelayVar	;wait
	
	rcall	DautoRefreshSet	;enable auto refresh (after each buffer write routine)
	
	
	;some text
	rcall	DclrAll
	ldiw	R30,Text2T*2
	rcall	DwrStrFLASH
	
	
	;some text
	ldi	R16,200
	rcall	DelayVar
	rcall	DclrAll
	ldiw	R30,Text3T*2
	rcall	DwrStrFLASH
	
	;BCD routines examples
	ldiw	R16,580
	rcall	bin2BCD16p5	;number to BCD
	
	ldi	R16,2
	rcall	DclrLine	;clear 2.line
	rcall	DwrBCD16z	;write num, with leading zeroes
	ldi	R16,200
	rcall	DelayVar	;wait 2s
	
	ldi	R16,2
	rcall	DclrLine	;clear 2.line
	rcall	DwrBCD16wz	;write num, white leading zeroes
	ldi	R16,200
	rcall	DelayVar	;wait 2s
	
	ldi	R16,2
	rcall	DclrLine	;clear 2.line
	rcall	DwrBCD16sz	;write num, skip leading zeroes
	ldi	R16,200
	rcall	DelayVar	;wait 2s
	
	
	;next text
	rcall	DclrAll
	ldiw	R30,Text4T*2
	rcall	DwrStrFLASH
	

	;flush USART udr
	in	R16,udr
	in	R16,udr
	
	
	;start terminal
Tloop:	rcall	COMrxByte	;wait for char from USART
	rcall	COMtxByte	;return received char

	cpi	R16,0x0D	;enter key to clear entire line
	brne	TermNoClrAll
	;clear line 
	ldi	R16,2
	rcall	DclrLine
	rcall	Drefresh
	rjmp	TloopE	
TermNoClrAll:	cpi	R16,0x08	;backspace key to erase one char
	brne	TermNoClrC
	;erase char
	ldsw	R28,LCDdataBufPtr
	cpiwh	R28,R16,LCDDMAdataBuf+LCDlineLen
	breq	TloopE
	;if not first char in the line
	ldi	R16,' '
	st	-Y,R16	;erase and decrement cursor pos
	stsw	LCDdataBufPtr,R28	;store cursor back
	rcall	Drefresh
	rjmp	TloopE
TermNoClrC:	;some printable char maybe
	rcall	DwrChar
	ldsw	R28,LCDdataBufPtr
	cpiwh	R28,R16,LCDDMAdataBuf+LCDlineLen*2
	brne	TloopE
	;line end
	sbiw	R28,1	;return 1 char back
	stsw	LCDdataBufPtr,R28	;store cursor back
TloopE:		
	rjmp	Tloop

	;text strings
Text1T:	.db	"HD44780 based LCD",0x00,"DMA routines",0xFF
Text2T:	.db	"by Stanislav Maslan",0x00,"s.maslan@seznam.cz",0xFF
Text3T:	.db	"BCD routines test:",0xFF
Text4T:	.db	"Terminal 19200bd:",0x00,0xFF	
	




;************************************************************************************************************************
; pseudo DMA LCD routines (written for ATmega32 and timer 2 compare match at T=50us)
; using some routines from direct HD44780 controlling routines
;  
; author: Stanislav Maslan, s.maslan@seznam.cz
;************************************************************************************************************************
;************************************************************************************************************************
;	LCD DMA refresh interruption routine (called by timer 2 compare match)
;	don't forget to place "rjmp INTLCDDMArefresh" to the timer 2 compare match interruption vector!  
;************************************************************************************************************************	
INTLCDDMArefresh:	pushw	R16
	pushw	R28
	in	R16,sreg
	push	R16
	;
	lds	R16,LCDDMAflags
	rjrc	R16,DposReset,INTLCDDMArefNNL
	;reset HD44780 cursor pos
	cbr	R16,(1<<DposReset)
	sts	LCDDMAflags,R16	;clear flag
	;reset pos
	cbi	LCDportRS,LRS	;instruction
	ldi	R16,0b10000000
	rcall	LCDbyteWriteI
	rjmp	INTLCDDMArefNL
	;no cursor reset
INTLCDDMArefNNL:	rjrc	R16,Dpos2ln,INTLCDDMArefNSL
	;break first line
	cbr	R16,(1<<Dpos2ln)
	sts	LCDDMAflags,R16	;clear flag
	;
	cbi	LCDportRS,LRS	;instruction
	ldi	R16,0b10000000+0x40
	rcall	LCDbyteWriteI
	rjmp	INTLCDDMArefNL
INTLCDDMArefNSL:	sbi	LCDportRS,LRS	;data
	ldsw	R28,LCDDMAdataBufPos
	ld	R16,Y+
	;high nibble
	rcall	LCDbyteWriteI
	;
	cpiwh	R28,R16,LCDDMAdataBuf+LCDlineLen
	brne	LCDDMAnslni
	;break line
	sbs	LCDDMAflags,Dpos2ln
	rjmp	INTLCDDMArefNC
LCDDMAnslni:	cpiwh	R28,R16,LCDDMAdataBuf+LCDlineLen*LCDlinesCnt
	brlo	INTLCDDMArefNC
	;refresh completed
	stsi	LCDDMAflags,0	;clear all flags
	cbii	timsk,ocie2	;disable refresh INT
	rjmp	INTLCDDMArefNC	;exit
	;refresh not completed
INTLCDDMArefNC:	stsw	LCDDMAdataBufPos,R28
INTLCDDMArefNL:	;reset timer
	outi	tcnt2,0	;clear timer
	outi	tifr,(1<<ocf2)	;clear INT request
	outi	sfior,(1<<psr2)	;and reset prescaler
	;
	pop	R16
	out	sreg,R16
	popw	R28
	popw	R16
	reti
	;write byte to LCD in 4-bit mode
LCDbyteWriteI:	push	R16
	swap	R16
	rcall	LCDnibOut
	rcall	LCDclock
	;low nibble
	pop	R16
	rcall	LCDnibOut
	rcall	LCDclock
	ret
;========================================================================================================================
;	Initialize LCD DMA refresh for R16 (1 or 2) lines LCD
;	byte TX period T=50us (don't forget set XTAL freq.!!!)
;========================================================================================================================
	;some LCD parameters
	.equ	LCDlineLen=20
	.equ	LCDlinesCnt=2
	;DMA flag byte bits
	.equ	DrefreshRequest=0
	.equ	DposReset=2
	.equ	Dpos2ln=3
	;LCD flag byte bits
	.equ	DautoRefresh=0

	
	;some variables for DMA refreshing
	.dseg
LCDDMAflags:	.byte	1	;LCD DMA refresh bit flags
LCDflags:	.byte	1	;LCD bit flags
LCDDMAdataBuf:	.byte	LCDlineLen*LCDlinesCnt	;character data buffer
LCDdataBufPtr:	.byte	2	;write buffer poitner
LCDDMAdataBufPos:	.byte	2	;buffer position
	.cseg

	;main init routine
Dinit:	rcall	LCDinit	;init LCD
	;fill entire character buffer with ' '
	rcall	DclrAll
	
	stsi	LCDflags,0	;clear all LCD bit flags
	stsi	LCDDMAflags,0	;clear refresh flags
	stsiw	LCDDMAdataBufPos,LCDDMAdataBuf	;reset buffer position
		
	outi	tccr2,(1<<cs21)|(1<<ctc2)	;timer 2 XTAL/8, CTC
	outi	ocr2,XTAL/8/20000-1	;f=20kHz (T=50us)
	ret
;========================================================================================================================
;	start LCD DMA refresh
;========================================================================================================================	
Drefresh:	in	R0,sreg
	cli
	outi	tcnt2,0
	outi	tifr,(1<<ocf2)
	outi	sfior,(1<<psr2)	;reset timer
	sbii	timsk,ocie2	;enable refresh IRQ
	stsiw	LCDDMAdataBufPos,LCDDMAdataBuf	;reset buffer position
	stsi	LCDDMAflags,(1<<DrefreshRequest)|(1<<DposReset)
	out	sreg,R0
	ret
;========================================================================================================================
;	wait for LCD DMA refresh complete 
;========================================================================================================================
DwaitRefresh:	lds	R16,LCDDMAflags
	andi	R16,(1<<DrefreshRequest)
	brne	DwaitRefresh	;wait if there is request for refresh or refresh in progress
	ret
;========================================================================================================================
;	test for LCD DMA refresh complete (T=1 if in progress (or i there is request for that), T=0 if done) 
;========================================================================================================================
DtestRefresh:	set		;in progress
	lds	R16,LCDDMAflags
	andi	R16,(1<<DrefreshRequest)
	brne	PC+2
	clt		;done
	ret
;========================================================================================================================
;	enable auto refresh
;========================================================================================================================
DautoRefreshSet:	sbs	LCDflags,DautoRefresh
	ret
;========================================================================================================================
;	disable auto refresh
;========================================================================================================================
DautoRefreshClr:	cbs	LCDflags,DautoRefresh
	ret
;========================================================================================================================
;	write R16 char
;========================================================================================================================
DwrChar:	ldsw	R26,LCDdataBufPtr
	st	X+,R16
	stsw	LCDdataBufPtr,R26
	;test autorefresh
	lds	R16,LCDflags
	rjrc	R16,DautoRefresh,PC+2
	rcall	Drefresh	;start refresh if its allowed
	ret
;========================================================================================================================
;	write Z pointed FLASH string into LCD data buffer 
;========================================================================================================================
DwrStrFLASH:	ldsw	R26,LCDdataBufPtr	;load cursor pos
	rjmp	DwrStrFLASHs
DwrStrFLASHL:	st	X+,R16	;write it to the LCD buffer
DwrStrFLASHs:	lpm	R16,Z+	;red char from FLASH
	cpi	R16,0	;break line?
	brne	DwrStrFLASHb
	ldiw	R26,LCDDMAdataBuf+LCDlineLen	;next line
	rjmp	DwrStrFLASHs
DwrStrFLASHb:	cpi	R16,0xFF	;string end?
	brne	DwrStrFLASHL
	stsw	LCDdataBufPtr,R26	;store cursor pos
	;test autorefresh
	lds	R16,LCDflags
	rjrc	R16,DautoRefresh,PC+2
	rcall	Drefresh	;start refresh if its allowed	
	ret
;========================================================================================================================
;	set cursor pos to R16
;	if your LCD size is 2x20: 
;	R16=10 - 1.line, 10.char
;	R16=20+5 - 2.line, 5.char
;========================================================================================================================	
DsetPos:	clr	R17
	subiw	R16,-(LCDDMAdataBuf)	;add buffer offset
	stsw	LCDdataBufPtr,R16
	ret
;========================================================================================================================
;	clear entire buffer, reset cursor to 0,0 (no auto refresh after this routine)
;========================================================================================================================
DclrAll:	;erase buffer (write ' ')
	ldiw	R28,LCDDMAdataBuf
	ldi	R24,LCDlineLen*LCDlinesCnt
	ldi	R16,' '
	st	Y+,R16
	dbnz	R24,PC-2
	stsiw	LCDdataBufPtr,LCDDMAdataBuf	;write cursor to 0,0
	ret
;========================================================================================================================
;	clear R16 line, reset cursor to its start
;========================================================================================================================
DclrLine:	dec	R16
	ldi	R17,20
	mul	R16,R17
	movw	R26,R0
	subiw	R26,-(LCDDMAdataBuf)	;line start address
	stsw	LCDdataBufPtr,R26	;store cursor pos
	;clear rest of this line
	ldi	R24,LCDlineLen
	ldi	R16,' '
	st	X+,R16
	dbnz	R24,PC-2
	ret	
;========================================================================================================================
;	write 16-bit BCD from R13:R14:R15 (packed format), use my speed optimized routines for bin/BCD conversion
;========================================================================================================================
	;with white leading zeroes
DwrBCD16wz:	clt
	clr	R0
	com	R0
	rjmp	DwrBCD16
	;with leading zeroes
DwrBCD16z:	set
	rjmp	DwrBCD16
	;skip leading zeroes
DwrBCD16sz:	clr	R0
	clt
	;common routine, don't call this directly!
DwrBCD16:	mov	R16,R15
	rcall	DwrCharBCD
	mov	R16,R14
	swap	R16
	rcall	DwrCharBCD
	mov	R16,R14
	rcall	DwrCharBCD
	mov	R16,R13
	swap	R16
	rcall	DwrCharBCD
	set
	mov	R16,R13
	rcall	DwrCharBCD
	ret
	;common write BCD num.
DwrCharBCD:	andi	R16,0b00001111	;mask low nibble
	brne	DwrCharBCDnz	;zero?
	brts	DwrCharBCDnz	;leading zero?
	;leading zero
	tst	R0	;space or nothing?
	breq	DwrCharBCDzc	;nothing
	ldi	R16,' '
	rjmp	DwrChar	;write number
DwrCharBCDnz:	subi	R16,-'0'	;to ASCII
	set
	rjmp	DwrChar	;write number
DwrCharBCDzc:	ret
	



;************************************************************************************************************************
; HD44780 LCD routines 4-wires interface, no RW pin used - constant delay 40us
; don't forget set correct XTAL freq somewhere
; don't forget init ddrx registers for E,RS and DATA pins immediately after MCU reset!!!
;
; author: Stanislav Maslan, s.maslan@seznam.cz 
;************************************************************************************************************************
	;pin/port definitions
	.equ	LCDportE=portd	;E pin port
	.equ	LCK=5	;pin?
	.equ	LCDportRS=portd	;RS pin port
	.equ	LRS=6	;pin?
	.equ	LCDportDATA=portc	;DATA pins port
	.equ	LCDportDATAnib=0	;low=0/high=1 nibble of the port

;------------------------------------------------------------------------------------------------------------------------
;	LCD initialization, call this routine first but generate at least 20ms delay from power on!
;	R16=1 one line, R16=2 two lines
;------------------------------------------------------------------------------------------------------------------------
LCDinit:	cbi	LCDportRS,LRS	;instructions
	push	R16
	;
	in	R0,sreg
	cli		;disable IRQ
	;back to 8-bit, if it was reset (brown out detector or ISP reset)
	ldi	R16,0b00000011	;to 8-bit mode
	rcall	LCDnibOut
	rcall	LCDclock	;clock
	rcall	LCDwait40	;wait 40us
	ldi	R16,0b00000000	;if it was 4-bit mode (after reset) complete instruction with low nibble 0b0000
	rcall	LCDnibOut	;if it was 8-bit mode, 0b00000000 instruction will be ignored
	rcall	LCDclock	;clock
	rcall	LCDwait164	;wait 1,64ms
	;now we are surely in 8-bit mode, init 4-bit mode
	ldi	R16,0b00000010
	rcall	LCDnibOut
	rcall	LCDclock	;clock
	rcall	LCDwait40	;wait 40us
	;
	out	sreg,R0
	;now select mode (R16=1 one line, R16=2 two lines)
	pop	R16
	dec	R16	;N--
	swap	R16
	lsr	R16	;N<<3
	andi	R16,(1<<3)
	ori	R16,0b00100000	;add init mask
	rcall	LCDwrInstr	;full init again into 4-bit mode, selected number of lines
	ldi	R16,0b00001100
	rcall	LCDwrInstr	;turn on display
	ldi	R16,0b00000110
	rcall	LCDwrInstr	;set cursor shift mode
	rcall	LCDclrAll	;clear all
	sbi	LCDportRS,LRS	;data	
	ret
;------------------------------------------------------------------------------------------------------------------------
;	LCD write 16-bit BCD from R13:R14:R15 (packed format), use my speed optimized routines
;------------------------------------------------------------------------------------------------------------------------
	;with white leading zeroes
LCDwrBCD16wz:	clt
	clr	R0
	com	R0
	rjmp	LCDwrBCD16
	;with leading zeroes
LCDwrBCD16z:	set
	rjmp	LCDwrBCD16
	;skip leading zeroes
LCDwrBCD16sz:	clr	R0
	clt
	;common routine, don't call this directly!
LCDwrBCD16:	mov	R16,R15
	rcall	LCDwrCharBCD
	mov	R16,R14
	swap	R16
	rcall	LCDwrCharBCD
	mov	R16,R14
	rcall	LCDwrCharBCD
	mov	R16,R13
	swap	R16
	rcall	LCDwrCharBCD
	set
	mov	R16,R13
	rcall	LCDwrCharBCD
	ret
	;common write BCD num.
LCDwrCharBCD:	andi	R16,0b00001111	;mask low nibble
	brne	LCDwrCharBCDnz	;zero?
	brts	LCDwrCharBCDnz	;leading zero?
	;leading zero
	tst	R0	;space or nothing?
	breq	LCDwrCharBCDzc	;nothing
	ldi	R16,' '
	rjmp	LCDwrChar	;write number
LCDwrCharBCDnz:	subi	R16,-'0'	;to ASCII
	set
	rjmp	LCDwrChar	;write number
LCDwrCharBCDzc:	ret
;------------------------------------------------------------------------------------------------------------------------
;	write Z pointed (in bytes!) 0xFF terminated string from FLASH, 0x00 line break
;	.db "text for 1st line",0x00,"text for second line",0xFF 
;------------------------------------------------------------------------------------------------------------------------
LCDwrStrFlash:	lpm	R16,Z+	;get char
	rcall	LCDwrChar
	lpm	R16,Z
	cpi	R16,0x00
	brne	LCDwrStrFlashNBL
	adiw	R30,1
	ldi	R16,0x40
	rcall	LCDsetPos
	rjmp	LCDwrStrFlash
LCDwrStrFlashNBL:	cpi	R16,0xFF	;str end?
	brne	LCDwrStrFlash	;next char
	ret
;------------------------------------------------------------------------------------------------------------------------
;	write Y pointed RAM array, R24=length
;------------------------------------------------------------------------------------------------------------------------
LCDwrStrRAM:	ld	R16,Y+
	rcall	LCDwrChar
	dbnz	R24,LCDwrStrRAM
	ret
;------------------------------------------------------------------------------------------------------------------------
;	set pos (R16):
;	1.line from 0x00
;	2.line from 0x40
;	3.line from 0x14
;	4.line from 0x54
;------------------------------------------------------------------------------------------------------------------------
LCDsetPos:	ori	R16,1<<7
	rjmp	LCDwrInstr
;------------------------------------------------------------------------------------------------------------------------
;	clr all
;------------------------------------------------------------------------------------------------------------------------
LCDclrAll:	ldi	R16,0b00000001
	rcall	LCDwrInstr
	rjmp	LCDwait164
;------------------------------------------------------------------------------------------------------------------------
;	clr R16 line, set cursor left on the line
;------------------------------------------------------------------------------------------------------------------------
LCDclrLine:	cpi	R16,1
	brne	LCDclrLineNo1
	ldi	R24,40	;40 chars
	ldi	R16,0x00	;1.line
	rjmp	LCDclrLineLC
LCDclrLineNo1:	cpi	R16,2
	brne	LCDclrLineNo2
	ldi	R24,40	;40 chars
	ldi	R16,0x40	;2.line
	rjmp	LCDclrLineLC
LCDclrLineNo2:	cpi	R16,3
	brne	LCDclrLineNo3
	ldi	R24,20	;20 chars
	ldi	R16,0x14	;3.line
	rjmp	LCDclrLineLC		
LCDclrLineNo3:	ldi	R24,20	;20 chars
	ldi	R16,0x54	;4.line
	rjmp	LCDclrLineLC
	;common clr routine
LCDclrLineLC:	push	R16
	rcall	LCDsetPos
LCDclrLineL:	ldi	R16,' '
	rcall	LCDwrChar
	dbnz	R24,LCDclrLineL
	pop	R16
	rcall	LCDsetPos
	ret

;------------------------------------------------------------------------------------------------------------------------
;	wait 40us
;------------------------------------------------------------------------------------------------------------------------	
LCDwait40:	ldi	R16,XTAL/74000
LCDwait40_1:	dbnz	R16,LCDwait40_1
	ret
;------------------------------------------------------------------------------------------------------------------------
;	wait 1.64ms
;------------------------------------------------------------------------------------------------------------------------	
LCDwait164:	ldi	R17,41	;41*40us
LCDwait164_2:	ldi	R16,XTAL/74000
LCDwait164_1:	dbnz	R16,LCDwait164_1
	dbnz	R17,LCDwait164_2
	ret
;------------------------------------------------------------------------------------------------------------------------
;	write R16 char, wait 40us
;------------------------------------------------------------------------------------------------------------------------
LCDwrChar:	sbi	LCDportRS,LRS	;data
	rcall	LCDwrByte
	rjmp	LCDwait40
;------------------------------------------------------------------------------------------------------------------------
;	write R16 instr, wait 40us
;------------------------------------------------------------------------------------------------------------------------
LCDwrInstr:	cbi	LCDportRS,LRS	;instructions
	rcall	LCDwrByte
	rjmp	LCDwait40
;------------------------------------------------------------------------------------------------------------------------
;	write R16 byte
;------------------------------------------------------------------------------------------------------------------------
LCDwrByte:	in	R0,sreg
	cli		;disable IRQ
	;high nibble
	push	R16
	swap	R16
	rcall	LCDnibOut
	rcall	LCDclock
	;low nibble
	pop	R16
	rcall	LCDnibOut
	rcall	LCDclock
	out	sreg,R0
	ret
;------------------------------------------------------------------------------------------------------------------------
;	write low nibble R16 to the LCD data port
;------------------------------------------------------------------------------------------------------------------------
LCDnibOut:	andi	R16,0b00001111	;mask low nibble
	.if	LCDportDATAnib==0	;low nibble?
	;low nibble
	in	R17,LCDportDATA
	andi	R17,0b11110000
	or	R17,R16	;to the port
	out	LCDportDATA,R17
	.elif	LCDportDATAnib==1	;high nibble?
	;high nibble
	swap	R16
	in	R17,LCDportDATA
	andi	R17,0b00001111
	or	R17,R16	;to the port
	out	LCDportDATA,R17
	.else
	.error "low=0 or high=1 nibble for LCD???"
	.endif
	ret
;------------------------------------------------------------------------------------------------------------------------
;	LCD clock
;------------------------------------------------------------------------------------------------------------------------	
LCDclock:	sbi	LCDportE,LCK
	nop
	nop
	cbi	LCDportE,LCK
	ret
	
	

;************************************************************************************************************************
;	delay routines
;************************************************************************************************************************
;------------------------------------------------------------------------------------------------------------------------
;	variable long delay (R16:R17 * 10ms)
;------------------------------------------------------------------------------------------------------------------------
DelayLongVar:	pushw	R18
DelayLongVar_3:	ldi	R19,250
DelayLongVar_2:	ldi	R18,XTAL/74000
DelayLongVar_1:	dbnz	R18,DelayLongVar_1
	dbnz	R19,DelayLongVar_2
	dbnzw	R16,DelayLongVar_3
	popw	R18
	ret
;------------------------------------------------------------------------------------------------------------------------
;	variable delay (R16 * 10ms)
;------------------------------------------------------------------------------------------------------------------------
DelayVar:	pushw	R18
DelayVar_3:	ldi	R19,250
DelayVar_2:	ldi	R18,XTAL/74000
DelayVar_1:	dbnz	R18,DelayVar_1
	dbnz	R19,DelayVar_2
	dbnz	R16,DelayVar_3
	popw	R18
	ret


	
;************************************************************************************************************************
;	math routines
;************************************************************************************************************************	
;------------------------------------------------------------------------------------------------------------------------
; bin2BCD16p5 - 16-bit binary to 5-digit packed BCD conversion (0..65535)
; prevod 16-bit binarniho cisla na peticiferne cislo v BCD
;
; BCD16res0:BCD16res1:BCD16res2 <- bin16r0:bin16r1
; BCD(0x01 0x23 0x45) <- bin(12345)
;      ^         ^
; BCD16res2  BCD16res0
;
; words:  29 + ret
; cycles: 29/180 + ret [min/max]
;
; author: Stanislav Maslan, s.maslan@seznam.cz
;------------------------------------------------------------------------------------------------------------------------
	.def	BCD16res0=R13
	.def	BCD16res1=R14
	.def	BCD16res2=R15
	.def	bin16r0=R16
	.def	bin16r1=R17
	.def	BB16_cnt1=R30
	.def	BB16_cnt2=R31
	;
bin2BCD16p5:	ser	BB16_cnt2
	mov	BCD16res2,BB16_cnt2
BB16p5_L10k:	inc	BCD16res2
	subi	bin16r0,low(10000)
	sbci	bin16r1,high(10000)
	brcc	BB16p5_L10k
	subi	bin16r0,low(-10000)
	sbci	bin16r1,high(-10000)
	ldi	BB16_cnt1,(256-16)
BB16p5_L1k:	subi	BB16_cnt1,(-16)
	subi	bin16r0,low(1000)
	sbci	bin16r1,high(1000)
	brcc	BB16p5_L1k
	subi	bin16r0,low(-1000)
	sbci	bin16r1,high(-1000)
	mov	BCD16res1,BB16_cnt2
BB16p5_L100:	inc	BCD16res1
	subi	bin16r0,low(100)
	sbci	bin16r1,high(100)
	brcc	BB16p5_L100
	subi	bin16r0,low(-100)
	or	BCD16res1,BB16_cnt1
	ldi	BB16_cnt1,(256-16)
BB16p5_L10:	subi	BB16_cnt1,(-16)
	subi	bin16r0,10
	brcc	BB16p5_L10
	subi	bin16r0,-10
	mov	BCD16res0,BB16_cnt1
	or	BCD16res0,bin16r0
	ret
	
	
	
	
;************************************************************************************************************************
;	USART routines
;************************************************************************************************************************	
;========================================================================================================================
;	USART write byte R16
;========================================================================================================================	
COMtxByte:	sbis	ucsra,udre
	rjmp	COMtxByte
	out	udr,R16
	ret
;========================================================================================================================
;	USART read byte to R16
;========================================================================================================================	
COMrxByte:	sbis	ucsra,rxc
	rjmp	COMrxByte	;wait for char from terminal
	in	R16,udr
	ret